@@ -3,6 +3,8 @@ import os.path
 import sys
 
 import charliecloud as ch
+import build_cache as bu
+
 
 ## Constants ##
 
@@ -16,24 +18,22 @@ manifests_internal = {
 }
 
 
+
 ## Main ##
 
 def main(cli):
    # Set things up.
-   ref = ch.Image_Ref(cli.image_ref)
+   src_ref = ch.Image_Ref(cli.source_ref)
+   dst_ref = src_ref if cli.dest_ref is None else ch.Image_Ref(cli.dest_ref)
    if (cli.parse_only):
-      print(ref.as_verbose_str)
+      print(src_ref.as_verbose_str)
       sys.exit(0)
-   image = ch.Image(ref, cli.image_dir)
-   ch.INFO("pulling image:    %s" % ref)
+   dst_img = ch.Image(dst_ref)
+   ch.INFO("pulling image:    %s" % src_ref)
+   if (src_ref != dst_ref):
+      ch.INFO("destination:      %s" % dst_ref)
    ch.INFO("requesting arch:  %s" % ch.arch)
-   if (cli.image_dir is not None):
-      ch.INFO("destination:      %s" % image.unpack_path)
-   else:
-      ch.VERBOSE("destination:      %s" % image.unpack_path)
-   pullet = Image_Puller(image, not cli.no_cache)
-   pullet.pull_to_unpacked(cli.last_layer)
-   pullet.done()
+   bu.cache.pull_eager(dst_img, src_ref, cli.last_layer)
    ch.done_notify()
 
 
@@ -46,15 +46,17 @@ class Image_Puller:
                 "image",
                 "layer_hashes",
                 "registry",
-                "use_cache")
+                "sid_input",
+                "src_ref")
 
-   def __init__(self, image, use_cache):
+   def __init__(self, image, src_ref):
       self.architectures = None
-      self.image = image
-      self.registry = ch.Registry_HTTP(image.ref)
       self.config_hash = None
+      self.image = image
       self.layer_hashes = None
-      self.use_cache = use_cache
+      self.registry = ch.Registry_HTTP(src_ref)
+      self.sid_input = None
+      self.src_ref = src_ref
 
    @property
    def config_path(self):
@@ -90,7 +92,7 @@ class Image_Puller:
          if (ch.arch != "yolo"):
             try:
                self.fatman_load()
-               if (ch.arch not in self.architectures):
+               if (not self.architectures.in_warn(ch.arch)):
                   ch.FATAL("requested arch unavailable: %s" % ch.arch,
                            ("available: %s"
                             % " ".join(sorted(self.architectures.keys()))))
@@ -106,12 +108,16 @@ class Image_Puller:
                            "consider --arch=yolo")
          # manifest
          self.manifest_load()
-      except ch.Not_In_Registry_Error:
-         ch.FATAL("not in registry: %s" % self.registry.ref)
+      except ch.Image_Unavailable_Error:
+         if (ch.user() == "qwofford"):
+            h = "Quincy, use --auth!!"
+         else:
+            h = "if your registry needs authentication, use --auth"
+         ch.FATAL("unauthorized or not in registry: %s" % self.registry.ref, h)
       # config
       ch.VERBOSE("config path: %s" % self.config_path)
       if (self.config_path is not None):
-         if (os.path.exists(self.config_path) and self.use_cache):
+         if (os.path.exists(self.config_path) and ch.dlcache_p):
             ch.INFO("config: using existing file")
          else:
             self.registry.blob_to_file(self.config_hash, self.config_path,
@@ -121,7 +127,7 @@ class Image_Puller:
          path = self.layer_path(lh)
          ch.VERBOSE("layer path: %s" % path)
          msg = "layer %d/%d: %s" % (i, len(self.layer_hashes), lh[:7])
-         if (os.path.exists(path) and self.use_cache):
+         if (os.path.exists(path) and ch.dlcache_p):
             ch.INFO("%s: using existing file" % msg)
          else:
             self.registry.blob_to_file(lh, path, "%s: downloading" % msg)
@@ -139,28 +145,26 @@ class Image_Puller:
       return (code, msg)
 
    def fatman_load(self):
-      """Load the fat manifest JSON file, downloading it first if needed. If
-         the image has a fat manifest, populate self.architectures; this may
-         be an empty dictionary if no valid architectures were found.
+      """Download the fat manifest and load it. If the image has a fat manifest
+         populate self.architectures; this may be an empty dictionary if no
+         valid architectures were found.
 
          Raises:
 
-           * Not_In_Registry_Error if the image does not exist.
+           * Image_Unavailable_Error if the image does not exist or we are not
+             authorized to have it.
 
            * No_Fatman_Error if the image exists but has no fat manifest,
              i.e., is architecture-unaware. In this case self.architectures is
              set to None."""
       self.architectures = None
-      if (str(self.image.ref) in manifests_internal):
+      if (str(self.src_ref) in manifests_internal):
          # cheat; internal manifest library matches every architecture
-         self.architectures = { ch.arch_host: None }
+         self.architectures = ch.Arch_Dict({ ch.arch_host: None })
          return
-      if (os.path.exists(self.fatman_path) and self.use_cache):
-         ch.INFO("manifest list: using existing file")
-      else:
-         # raises Not_In_Registry_Error if needed
-         self.registry.fatman_to_file(self.fatman_path,
-                                      "manifest list: downloading")
+      # raises Image_Unavailable_Error if needed
+      self.registry.fatman_to_file(self.fatman_path,
+                                   "manifest list: downloading")
       fm = ch.json_from_file(self.fatman_path, "fat manifest")
       if ("layers" in fm or "fsLayers" in fm):
          # FIXME (issue #1101): If it's a v2 manifest we could use it instead
@@ -175,7 +179,7 @@ class Image_Puller:
             return
          else:
             ch.FATAL("manifest list: error: %s" % msg)
-      self.architectures = dict()
+      self.architectures = ch.Arch_Dict()
       if ("manifests" not in fm):
          ch.FATAL("manifest list has no key 'manifests'")
       for m in fm["manifests"]:
@@ -199,8 +203,8 @@ class Image_Puller:
       return ch.storage.download_cache // (layer_hash + ".tar.gz")
 
    def manifest_load(self):
-      """Download the manifest file if needed, parse it, and set
-         self.config_hash and self.layer_hashes. If the image does not exist,
+      """Download the manifest file, parse it, and set self.config_hash and
+         self.layer_hashes. If the image does not exist,
          exit with error."""
       def bad_key(key):
          ch.FATAL("manifest: %s: no key: %s" % (self.manifest_path, key))
@@ -209,21 +213,18 @@ class Image_Puller:
       # obtain the manifest
       try:
          # internal manifest library, e.g. for "FROM scratch"
-         manifest = manifests_internal[str(self.image.ref)]
+         manifest = manifests_internal[str(self.src_ref)]
          ch.INFO("manifest: using internal library")
       except KeyError:
-         # download the file if needed, then parse it
+         # download the file and parse it
          if (ch.arch == "yolo" or self.architectures is None):
             digest = None
          else:
             digest = self.architectures[ch.arch]
          ch.DEBUG("manifest digest: %s" % digest)
-         if (os.path.exists(self.manifest_path) and self.use_cache):
-            ch.INFO("manifest: using existing file")
-         else:
-            self.registry.manifest_to_file(self.manifest_path,
-                                           "manifest: downloading",
-                                           digest=digest)
+         self.registry.manifest_to_file(self.manifest_path,
+                                        "manifest: downloading",
+                                        digest=digest)
          manifest = ch.json_from_file(self.manifest_path, "manifest")
       # validate schema version
       try:
@@ -263,6 +264,9 @@ class Image_Puller:
          self.layer_hashes.append(ch.digest_trim(i[key2]))
       if (version == 1):
          self.layer_hashes.reverse()
+      # Remember State_ID input. We can't rely on the manifest existing in
+      # serialized form (e.g. for internal manifests), so re-serialize.
+      self.sid_input = json.dumps(manifest, sort_keys=True)
 
    def manifest_digest_by_arch(self):
       "Return skinny manifest digest for target architecture."
@@ -288,9 +292,7 @@ class Image_Puller:
                   'try "ch-image list IMAGE_REF"')
       return digest
 
-   def pull_to_unpacked(self, last_layer=None):
-      "Pull and flatten image."
-      self.download()
+   def unpack(self, last_layer=None):
       layer_paths = [self.layer_path(h) for h in self.layer_hashes]
       self.image.unpack(layer_paths, last_layer)
       self.image.metadata_replace(self.config_path)
