@@ -115,11 +115,17 @@ class Image_Pusher:
       # Prepare metadata.
       ch.INFO("preparing metadata")
       self.image.metadata_load()
-      hist = self.image.metadata["history"]
-      # Some registries, e.g., Quay, use history metadata for simple sanity
-      # checks. For example, when an image's number of "empty_layer" history
-      # entries doesn't match the number of layers being uploaded, Quay will
-      # reject the image upload.
+      # Environment. Note that this is *not* a dictionary for some reason but
+      # a list of name/value pairs separated by equals [1], with no quoting.
+      #
+      # [1]: https://github.com/opencontainers/image-spec/blob/main/config.md
+      config['config']['Env'] = ["%s=%s" % (k, v)
+                                 for k, v
+                                 in self.image.metadata.get("env", {}).items()]
+      # History. Some registries, e.g., Quay, use history metadata for simple
+      # sanity checks. For example, when an image's number of "empty_layer"
+      # history entries doesn't match the number of layers being uploaded,
+      # Quay will reject the image upload.
       #
       # This type of error checking is odd as the empty_layer key is optional
       # (https://github.com/opencontainers/image-spec/blob/main/config.md).
@@ -128,6 +134,7 @@ class Image_Pusher:
       # the total number of non-empty layers always totals one (1). To do this
       # we iterate over the history entires backward searching for the first
       # non-empty entry and preserve it; all others are set to empty.
+      hist = self.image.metadata["history"]
       non_empty_winner = None
       for i in range(len(hist) - 1, -1, -1):
          if (   "empty_layer" not in hist[i].keys()
@@ -140,6 +147,7 @@ class Image_Pusher:
          if (i != non_empty_winner):
             hist[i]["empty_layer"] = True
       config["history"] = hist
+      # Pack it up to go.
       config_bytes = json.dumps(config, indent=2).encode("UTF-8")
       config_hash = ch.bytes_hash(config_bytes)
       manifest["config"]["size"] = len(config_bytes)
